@@ -1,7 +1,7 @@
 /*
  * Copyright © 2009 CNRS
- * Copyright © 2009-2014 Inria.  All rights reserved.
- * Copyright © 2009-2011 Université Bordeaux 1
+ * Copyright © 2009-2015 Inria.  All rights reserved.
+ * Copyright © 2009-2011 Université Bordeaux
  * Copyright © 2009-2011 Cisco Systems, Inc.  All rights reserved.
  * See COPYING in top-level directory.
  */
@@ -19,7 +19,7 @@ hwloc__xml_verbose(void)
   static int first = 1;
   static int verbose = 0;
   if (first) {
-    char *env = getenv("HWLOC_XML_VERBOSE");
+    const char *env = getenv("HWLOC_XML_VERBOSE");
     if (env)
       verbose = atoi(env);
     first = 0;
@@ -27,6 +27,34 @@ hwloc__xml_verbose(void)
   return verbose;
 }
 
+static int
+hwloc_nolibxml_import(void)
+{
+  static int first = 1;
+  static int nolibxml = 0;
+  if (first) {
+    const char *env = getenv("HWLOC_NO_LIBXML_IMPORT");
+    if (env)
+      nolibxml = atoi(env);
+    first = 0;
+  }
+  return nolibxml;
+}
+
+static int
+hwloc_nolibxml_export(void)
+{
+  static int first = 1;
+  static int nolibxml = 0;
+  if (first) {
+    const char *env = getenv("HWLOC_NO_LIBXML_EXPORT");
+    if (env)
+      nolibxml = atoi(env);
+    first = 0;
+  }
+  return nolibxml;
+}
+
 /*********************************
  ********* XML callbacks *********
  *********************************/
@@ -51,7 +79,7 @@ hwloc_xml_callbacks_reset(void)
 {
   hwloc_nolibxml_callbacks = NULL;
   hwloc_libxml_callbacks = NULL;
-}			       
+}
 
 /************************************************
  ********* XML import (common routines) *********
@@ -59,7 +87,8 @@ hwloc_xml_callbacks_reset(void)
 
 static void
 hwloc__xml_import_object_attr(struct hwloc_topology *topology __hwloc_attribute_unused, struct hwloc_obj *obj,
-			      const char *name, const char *value)
+			      const char *name, const char *value,
+			      hwloc__xml_import_state_t state)
 {
   if (!strcmp(name, "type")) {
     /* already handled */
@@ -99,7 +128,8 @@ hwloc__xml_import_object_attr(struct hwloc_topology *topology __hwloc_attribute_
     if (obj->type == HWLOC_OBJ_CACHE)
       obj->attr->cache.size = lvalue;
     else if (hwloc__xml_verbose())
-      fprintf(stderr, "ignoring cache_size attribute for non-cache object type\n");
+      fprintf(stderr, "%s: ignoring cache_size attribute for non-cache object type\n",
+	      state->global->msgprefix);
   }
 
   else if (!strcmp(name, "cache_linesize")) {
@@ -107,7 +137,8 @@ hwloc__xml_import_object_attr(struct hwloc_topology *topology __hwloc_attribute_
     if (obj->type == HWLOC_OBJ_CACHE)
       obj->attr->cache.linesize = lvalue;
     else if (hwloc__xml_verbose())
-      fprintf(stderr, "ignoring cache_linesize attribute for non-cache object type\n");
+      fprintf(stderr, "%s: ignoring cache_linesize attribute for non-cache object type\n",
+	      state->global->msgprefix);
   }
 
   else if (!strcmp(name, "cache_associativity")) {
@@ -115,7 +146,8 @@ hwloc__xml_import_object_attr(struct hwloc_topology *topology __hwloc_attribute_
     if (obj->type == HWLOC_OBJ_CACHE)
       obj->attr->cache.associativity = lvalue;
     else if (hwloc__xml_verbose())
-      fprintf(stderr, "ignoring cache_associativity attribute for non-cache object type\n");
+      fprintf(stderr, "%s: ignoring cache_associativity attribute for non-cache object type\n",
+	      state->global->msgprefix);
   }
 
   else if (!strcmp(name, "cache_type")) {
@@ -126,9 +158,11 @@ hwloc__xml_import_object_attr(struct hwloc_topology *topology __hwloc_attribute_
 	  || lvalue == HWLOC_OBJ_CACHE_INSTRUCTION)
 	obj->attr->cache.type = (hwloc_obj_cache_type_t) lvalue;
       else
-	fprintf(stderr, "ignoring invalid cache_type attribute %ld\n", lvalue);
+	fprintf(stderr, "%s: ignoring invalid cache_type attribute %ld\n",
+		state->global->msgprefix, lvalue);
     } else if (hwloc__xml_verbose())
-      fprintf(stderr, "ignoring cache_type attribute for non-cache object type\n");
+      fprintf(stderr, "%s: ignoring cache_type attribute for non-cache object type\n",
+	      state->global->msgprefix);
   }
 
   else if (!strcmp(name, "local_memory"))
@@ -148,7 +182,8 @@ hwloc__xml_import_object_attr(struct hwloc_topology *topology __hwloc_attribute_
 	break;
       default:
 	if (hwloc__xml_verbose())
-	  fprintf(stderr, "ignoring depth attribute for object type without depth\n");
+	  fprintf(stderr, "%s: ignoring depth attribute for object type without depth\n",
+		  state->global->msgprefix);
 	break;
     }
   }
@@ -161,7 +196,8 @@ hwloc__xml_import_object_attr(struct hwloc_topology *topology __hwloc_attribute_
       if (sscanf(value, "%04x:%02x:%02x.%01x",
 		 &domain, &bus, &dev, &func) != 4) {
 	if (hwloc__xml_verbose())
-	  fprintf(stderr, "ignoring invalid pci_busid format string %s\n", value);
+	  fprintf(stderr, "%s: ignoring invalid pci_busid format string %s\n",
+		  state->global->msgprefix, value);
       } else {
 	obj->attr->pcidev.domain = domain;
 	obj->attr->pcidev.bus = bus;
@@ -172,7 +208,8 @@ hwloc__xml_import_object_attr(struct hwloc_topology *topology __hwloc_attribute_
     }
     default:
       if (hwloc__xml_verbose())
-	fprintf(stderr, "ignoring pci_busid attribute for non-PCI object\n");
+	fprintf(stderr, "%s: ignoring pci_busid attribute for non-PCI object\n",
+		state->global->msgprefix);
       break;
     }
   }
@@ -185,7 +222,8 @@ hwloc__xml_import_object_attr(struct hwloc_topology *topology __hwloc_attribute_
       if (sscanf(value, "%04x [%04x:%04x] [%04x:%04x] %02x",
 		 &classid, &vendor, &device, &subvendor, &subdevice, &revision) != 6) {
 	if (hwloc__xml_verbose())
-	  fprintf(stderr, "ignoring invalid pci_type format string %s\n", value);
+	  fprintf(stderr, "%s: ignoring invalid pci_type format string %s\n",
+		  state->global->msgprefix, value);
       } else {
 	obj->attr->pcidev.class_id = classid;
 	obj->attr->pcidev.vendor_id = vendor;
@@ -198,7 +236,8 @@ hwloc__xml_import_object_attr(struct hwloc_topology *topology __hwloc_attribute_
     }
     default:
       if (hwloc__xml_verbose())
-	fprintf(stderr, "ignoring pci_type attribute for non-PCI object\n");
+	fprintf(stderr, "%s: ignoring pci_type attribute for non-PCI object\n",
+		state->global->msgprefix);
       break;
     }
   }
@@ -212,7 +251,8 @@ hwloc__xml_import_object_attr(struct hwloc_topology *topology __hwloc_attribute_
     }
     default:
       if (hwloc__xml_verbose())
-	fprintf(stderr, "ignoring pci_link_speed attribute for non-PCI object\n");
+	fprintf(stderr, "%s: ignoring pci_link_speed attribute for non-PCI object\n",
+		state->global->msgprefix);
       break;
     }
   }
@@ -223,7 +263,8 @@ hwloc__xml_import_object_attr(struct hwloc_topology *topology __hwloc_attribute_
       unsigned upstream_type, downstream_type;
       if (sscanf(value, "%u-%u", &upstream_type, &downstream_type) != 2) {
 	if (hwloc__xml_verbose())
-	  fprintf(stderr, "ignoring invalid bridge_type format string %s\n", value);
+	  fprintf(stderr, "%s: ignoring invalid bridge_type format string %s\n",
+		  state->global->msgprefix, value);
       } else {
 	obj->attr->bridge.upstream_type = (hwloc_obj_bridge_type_t) upstream_type;
 	obj->attr->bridge.downstream_type = (hwloc_obj_bridge_type_t) downstream_type;
@@ -232,7 +273,8 @@ hwloc__xml_import_object_attr(struct hwloc_topology *topology __hwloc_attribute_
     }
     default:
       if (hwloc__xml_verbose())
-	fprintf(stderr, "ignoring bridge_type attribute for non-bridge object\n");
+	fprintf(stderr, "%s: ignoring bridge_type attribute for non-bridge object\n",
+		state->global->msgprefix);
       break;
     }
   }
@@ -244,7 +286,8 @@ hwloc__xml_import_object_attr(struct hwloc_topology *topology __hwloc_attribute_
       if (sscanf(value, "%04x:[%02x-%02x]",
 		 &domain, &secbus, &subbus) != 3) {
 	if (hwloc__xml_verbose())
-	  fprintf(stderr, "ignoring invalid bridge_pci format string %s\n", value);
+	  fprintf(stderr, "%s: ignoring invalid bridge_pci format string %s\n",
+		  state->global->msgprefix, value);
       } else {
 	obj->attr->bridge.downstream.pci.domain = domain;
 	obj->attr->bridge.downstream.pci.secondary_bus = secbus;
@@ -254,7 +297,8 @@ hwloc__xml_import_object_attr(struct hwloc_topology *topology __hwloc_attribute_
     }
     default:
       if (hwloc__xml_verbose())
-	fprintf(stderr, "ignoring bridge_pci attribute for non-bridge object\n");
+	fprintf(stderr, "%s: ignoring bridge_pci attribute for non-bridge object\n",
+		state->global->msgprefix);
       break;
     }
   }
@@ -265,14 +309,16 @@ hwloc__xml_import_object_attr(struct hwloc_topology *topology __hwloc_attribute_
       unsigned osdev_type;
       if (sscanf(value, "%u", &osdev_type) != 1) {
 	if (hwloc__xml_verbose())
-	  fprintf(stderr, "ignoring invalid osdev_type format string %s\n", value);
+	  fprintf(stderr, "%s: ignoring invalid osdev_type format string %s\n",
+		  state->global->msgprefix, value);
       } else
 	obj->attr->osdev.type = (hwloc_obj_osdev_type_t) osdev_type;
       break;
     }
     default:
       if (hwloc__xml_verbose())
-	fprintf(stderr, "ignoring osdev_type attribute for non-osdev object\n");
+	fprintf(stderr, "%s: ignoring osdev_type attribute for non-osdev object\n",
+		state->global->msgprefix);
       break;
     }
   }
@@ -284,10 +330,10 @@ hwloc__xml_import_object_attr(struct hwloc_topology *topology __hwloc_attribute_
    * deprecated (from 1.0)
    */
   else if (!strcmp(name, "dmi_board_vendor")) {
-    hwloc_obj_add_info(obj, "DMIBoardVendor", strdup(value));
+    hwloc_obj_add_info(obj, "DMIBoardVendor", value);
   }
   else if (!strcmp(name, "dmi_board_name")) {
-    hwloc_obj_add_info(obj, "DMIBoardName", strdup(value));
+    hwloc_obj_add_info(obj, "DMIBoardName", value);
   }
 
   /*************************
@@ -299,21 +345,22 @@ hwloc__xml_import_object_attr(struct hwloc_topology *topology __hwloc_attribute_
       case HWLOC_OBJ_CACHE:
 	obj->attr->cache.size = lvalue << 10;
 	break;
-      case HWLOC_OBJ_NODE:
+      case HWLOC_OBJ_NUMANODE:
       case HWLOC_OBJ_MACHINE:
       case HWLOC_OBJ_SYSTEM:
 	obj->memory.local_memory = lvalue << 10;
 	break;
       default:
 	if (hwloc__xml_verbose())
-	  fprintf(stderr, "ignoring memory_kB attribute for object type without memory\n");
+	  fprintf(stderr, "%s: ignoring memory_kB attribute for object type without memory\n",
+		  state->global->msgprefix);
 	break;
     }
   }
   else if (!strcmp(name, "huge_page_size_kB")) {
     unsigned long lvalue = strtoul(value, NULL, 10);
     switch (obj->type) {
-      case HWLOC_OBJ_NODE:
+      case HWLOC_OBJ_NUMANODE:
       case HWLOC_OBJ_MACHINE:
       case HWLOC_OBJ_SYSTEM:
 	if (!obj->memory.page_types) {
@@ -324,14 +371,15 @@ hwloc__xml_import_object_attr(struct hwloc_topology *topology __hwloc_attribute_
 	break;
       default:
 	if (hwloc__xml_verbose())
-	  fprintf(stderr, "ignoring huge_page_size_kB attribute for object type without huge pages\n");
+	  fprintf(stderr, "%s: ignoring huge_page_size_kB attribute for object type without huge pages\n",
+		  state->global->msgprefix);
 	break;
     }
   }
   else if (!strcmp(name, "huge_page_free")) {
     unsigned long lvalue = strtoul(value, NULL, 10);
     switch (obj->type) {
-      case HWLOC_OBJ_NODE:
+      case HWLOC_OBJ_NUMANODE:
       case HWLOC_OBJ_MACHINE:
       case HWLOC_OBJ_SYSTEM:
 	if (!obj->memory.page_types) {
@@ -342,7 +390,8 @@ hwloc__xml_import_object_attr(struct hwloc_topology *topology __hwloc_attribute_
 	break;
       default:
 	if (hwloc__xml_verbose())
-	  fprintf(stderr, "ignoring huge_page_free attribute for object type without huge pages\n");
+	  fprintf(stderr, "%s: ignoring huge_page_free attribute for object type without huge pages\n",
+		  state->global->msgprefix);
 	break;
     }
   }
@@ -353,7 +402,8 @@ hwloc__xml_import_object_attr(struct hwloc_topology *topology __hwloc_attribute_
 
 
   else if (hwloc__xml_verbose())
-    fprintf(stderr, "ignoring unknown object attribute %s\n", name);
+    fprintf(stderr, "%s: ignoring unknown object attribute %s\n",
+	    state->global->msgprefix, name);
 }
 
 
@@ -366,7 +416,7 @@ hwloc__xml_import_info(hwloc_topology_t topology __hwloc_attribute_unused, hwloc
 
   while (1) {
     char *attrname, *attrvalue;
-    if (state->next_attr(state, &attrname, &attrvalue) < 0)
+    if (state->global->next_attr(state, &attrname, &attrvalue) < 0)
       break;
     if (!strcmp(attrname, "name"))
       infoname = attrvalue;
@@ -380,7 +430,7 @@ hwloc__xml_import_info(hwloc_topology_t topology __hwloc_attribute_unused, hwloc
     /* empty strings are ignored by libxml */
     hwloc_obj_add_info(obj, infoname, infovalue ? infovalue : "");
 
-  return state->close_tag(state);
+  return state->global->close_tag(state);
 }
 
 static int
@@ -391,7 +441,7 @@ hwloc__xml_import_pagetype(hwloc_topology_t topology __hwloc_attribute_unused, h
 
   while (1) {
     char *attrname, *attrvalue;
-    if (state->next_attr(state, &attrname, &attrvalue) < 0)
+    if (state->global->next_attr(state, &attrname, &attrvalue) < 0)
       break;
     if (!strcmp(attrname, "size"))
       size = strtoull(attrvalue, NULL, 10);
@@ -409,7 +459,7 @@ hwloc__xml_import_pagetype(hwloc_topology_t topology __hwloc_attribute_unused, h
     obj->memory.page_types[idx].count = count;
   }
 
-  return state->close_tag(state);
+  return state->global->close_tag(state);
 }
 
 static int
@@ -424,7 +474,7 @@ hwloc__xml_import_distances(struct hwloc_xml_backend_data_s *data,
 
   while (1) {
     char *attrname, *attrvalue;
-    if (state->next_attr(state, &attrname, &attrvalue) < 0)
+    if (state->global->next_attr(state, &attrname, &attrvalue) < 0)
       break;
     if (!strcmp(attrname, "nbobjs"))
       nbobjs = strtoul(attrvalue, NULL, 10);
@@ -453,7 +503,7 @@ hwloc__xml_import_distances(struct hwloc_xml_backend_data_s *data,
       char *attrname, *attrvalue;
       float val;
 
-      ret = state->find_child(state, &childstate, &tag);
+      ret = state->global->find_child(state, &childstate, &tag);
       if (ret <= 0 || strcmp(tag, "latency")) {
 	/* a latency child is needed */
 	free(distances->distances.latency);
@@ -461,7 +511,7 @@ hwloc__xml_import_distances(struct hwloc_xml_backend_data_s *data,
 	return -1;
       }
 
-      ret = state->next_attr(&childstate, &attrname, &attrvalue);
+      ret = state->global->next_attr(&childstate, &attrname, &attrvalue);
       if (ret < 0 || strcmp(attrname, "value")) {
 	free(distances->distances.latency);
 	free(distances);
@@ -473,11 +523,11 @@ hwloc__xml_import_distances(struct hwloc_xml_backend_data_s *data,
       if (val > latmax)
 	latmax = val;
 
-      ret = state->close_tag(&childstate);
+      ret = state->global->close_tag(&childstate);
       if (ret < 0)
 	return -1;
 
-      state->close_child(&childstate);
+      state->global->close_child(&childstate);
     }
 
     distances->distances.latency_max = latmax;
@@ -490,7 +540,7 @@ hwloc__xml_import_distances(struct hwloc_xml_backend_data_s *data,
     distances->next = NULL;
   }
 
-  return state->close_tag(state);
+  return state->global->close_tag(state);
 }
 
 static int
@@ -503,7 +553,7 @@ hwloc__xml_import_userdata(hwloc_topology_t topology __hwloc_attribute_unused, h
 
   while (1) {
     char *attrname, *attrvalue;
-    if (state->next_attr(state, &attrname, &attrvalue) < 0)
+    if (state->global->next_attr(state, &attrname, &attrvalue) < 0)
       break;
     if (!strcmp(attrname, "length"))
       length = strtoul(attrvalue, NULL, 10);
@@ -521,7 +571,7 @@ hwloc__xml_import_userdata(hwloc_topology_t topology __hwloc_attribute_unused, h
     if (encoded) {
       char *encoded_buffer;
       size_t encoded_length = 4*((length+2)/3);
-      ret = state->get_content(state, &encoded_buffer, encoded_length);
+      ret = state->global->get_content(state, &encoded_buffer, encoded_length);
       if (ret < 0)
         return -1;
       if (ret) {
@@ -537,14 +587,14 @@ hwloc__xml_import_userdata(hwloc_topology_t topology __hwloc_attribute_unused, h
       }
     } else {
       char *buffer;
-      ret = state->get_content(state, &buffer, length);
+      ret = state->global->get_content(state, &buffer, length);
       if (ret < 0)
         return -1;
       topology->userdata_import_cb(topology, obj, name, buffer, length);
     }
-    state->close_content(state);
+    state->global->close_content(state);
   }
-  return state->close_tag(state);
+  return state->global->close_tag(state);
 }
 
 static int
@@ -553,24 +603,78 @@ hwloc__xml_import_object(hwloc_topology_t topology,
 			 hwloc_obj_t obj,
 			 hwloc__xml_import_state_t state)
 {
+  hwloc_obj_t parent = obj->parent;
+
   /* process attributes */
   while (1) {
     char *attrname, *attrvalue;
-    if (state->next_attr(state, &attrname, &attrvalue) < 0)
+    if (state->global->next_attr(state, &attrname, &attrvalue) < 0)
       break;
     if (!strcmp(attrname, "type")) {
       if (hwloc_obj_type_sscanf(attrvalue, &obj->type, NULL, NULL, 0) < 0)
-        return -1;
+	goto error;
     } else {
       /* type needed first */
       if (obj->type == (hwloc_obj_type_t)-1)
-        return -1;
-      hwloc__xml_import_object_attr(topology, obj, attrname, attrvalue);
+	goto error;
+      hwloc__xml_import_object_attr(topology, obj, attrname, attrvalue, state);
     }
   }
 
-  if (obj->parent) {
+  if (parent) {
     /* root->parent is NULL, and root is already inserted */
+
+    /* warn if inserting out-of-order */
+    if (parent->cpuset) { /* don't compare children if multinode parent */
+      hwloc_obj_t *current;
+      for (current = &parent->first_child; *current; current = &(*current)->next_sibling) {
+	hwloc_bitmap_t curcpuset = (*current)->cpuset;
+	if (obj->cpuset && (!curcpuset || hwloc__object_cpusets_compare_first(obj, *current) < 0)) {
+	  static int reported = 0;
+	  if (!reported && !hwloc_hide_errors()) {
+	    char *progname = hwloc_progname(topology);
+	    const char *origversion = hwloc_obj_get_info_by_name(topology->levels[0][0], "hwlocVersion");
+	    const char *origprogname = hwloc_obj_get_info_by_name(topology->levels[0][0], "ProcessName");
+	    char *c1, *cc1, t1[64];
+	    char *c2 = NULL, *cc2 = NULL, t2[64];
+	    hwloc_bitmap_asprintf(&c1, obj->cpuset);
+	    hwloc_bitmap_asprintf(&cc1, obj->complete_cpuset);
+	    hwloc_obj_type_snprintf(t1, sizeof(t1), obj, 0);
+	    if (curcpuset)
+	      hwloc_bitmap_asprintf(&c2, curcpuset);
+	    if ((*current)->complete_cpuset)
+	      hwloc_bitmap_asprintf(&cc2, (*current)->complete_cpuset);
+	    hwloc_obj_type_snprintf(t2, sizeof(t2), *current, 0);
+	    fprintf(stderr, "****************************************************************************\n");
+	    fprintf(stderr, "* hwloc has encountered an out-of-order XML topology load.\n");
+	    fprintf(stderr, "* Object %s cpuset %s complete %s\n",
+		    t1, c1, cc1);
+	    fprintf(stderr, "* was inserted after object %s with %s and %s.\n",
+		    t2, c2 ? c2 : "none", cc2 ? cc2 : "none");
+	    fprintf(stderr, "* The error occured in hwloc %s inside process `%s', while\n",
+		    HWLOC_VERSION,
+		    progname ? progname : "<unknown>");
+	    if (origversion || origprogname)
+	      fprintf(stderr, "* the input XML was generated by hwloc %s inside process `%s'.\n",
+		      origversion ? origversion : "(unknown version)",
+		      origprogname ? origprogname : "<unknown>");
+	    else
+	      fprintf(stderr, "* the input XML was generated by an unspecified ancient hwloc release.\n");
+	    fprintf(stderr, "* Please check that your input topology XML file is valid.\n");
+	    fprintf(stderr, "****************************************************************************\n");
+	    free(c1);
+	    free(cc1);
+	    if (c2)
+	      free(c2);
+	    if (cc2)
+	      free(cc2);
+	    free(progname);
+	    reported = 1;
+	  }
+	}
+      }
+    }
+
     hwloc_insert_object_by_parent(topology, obj->parent /* filled by the caller */, obj);
     /* insert_object_by_parent() doesn't merge during insert, so obj is still valid */
   }
@@ -581,9 +685,9 @@ hwloc__xml_import_object(hwloc_topology_t topology,
     char *tag;
     int ret;
 
-    ret = state->find_child(state, &childstate, &tag);
+    ret = state->global->find_child(state, &childstate, &tag);
     if (ret < 0)
-      return -1;
+      goto error;
     if (!ret)
       break;
 
@@ -603,12 +707,16 @@ hwloc__xml_import_object(hwloc_topology_t topology,
       ret = -1;
 
     if (ret < 0)
-      return ret;
+      goto error;
 
-    state->close_child(&childstate);
+    state->global->close_child(&childstate);
   }
 
-  return state->close_tag(state);
+  return state->global->close_tag(state);
+
+ error:
+  hwloc_free_unlinked_object(obj);
+  return -1;
 }
 
 static int
@@ -627,7 +735,7 @@ hwloc__xml_import_diff_one(hwloc__xml_import_state_t state,
 
   while (1) {
     char *attrname, *attrvalue;
-    if (state->next_attr(state, &attrname, &attrvalue) < 0)
+    if (state->global->next_attr(state, &attrname, &attrvalue) < 0)
       break;
     if (!strcmp(attrname, "type"))
       type_s = attrvalue;
@@ -645,8 +753,12 @@ hwloc__xml_import_diff_one(hwloc__xml_import_state_t state,
       obj_attr_oldvalue_s = attrvalue;
     else if (!strcmp(attrname, "obj_attr_newvalue"))
       obj_attr_newvalue_s = attrvalue;
-    else
+    else {
+      if (hwloc__xml_verbose())
+	fprintf(stderr, "%s: ignoring unknown diff attribute %s\n",
+		state->global->msgprefix, attrname);
       return -1;
+    }
   }
 
   if (type_s) {
@@ -659,17 +771,29 @@ hwloc__xml_import_diff_one(hwloc__xml_import_state_t state,
       hwloc_topology_diff_t diff;
 
       /* obj_attr mandatory generic attributes */
-      if (!obj_depth_s || !obj_index_s || !obj_attr_type_s)
+      if (!obj_depth_s || !obj_index_s || !obj_attr_type_s) {
+	if (hwloc__xml_verbose())
+	  fprintf(stderr, "%s: missing mandatory obj attr generic attributes\n",
+		  state->global->msgprefix);
 	break;
+      }
 
       /* obj_attr mandatory attributes common to all subtypes */
-      if (!obj_attr_oldvalue_s || !obj_attr_newvalue_s)
+      if (!obj_attr_oldvalue_s || !obj_attr_newvalue_s) {
+	if (hwloc__xml_verbose())
+	  fprintf(stderr, "%s: missing mandatory obj attr value attributes\n",
+		  state->global->msgprefix);
 	break;
+      }
 
       /* mandatory attributes for obj_attr_info subtype */
       obj_attr_type = atoi(obj_attr_type_s);
-      if (obj_attr_type == HWLOC_TOPOLOGY_DIFF_OBJ_ATTR_INFO && !obj_attr_name_s)
+      if (obj_attr_type == HWLOC_TOPOLOGY_DIFF_OBJ_ATTR_INFO && !obj_attr_name_s) {
+	if (hwloc__xml_verbose())
+	  fprintf(stderr, "%s: missing mandatory obj attr info name attribute\n",
+		  state->global->msgprefix);
 	break;
+      }
 
       /* now we know we have everything we need */
       diff = malloc(sizeof(*diff));
@@ -705,7 +829,7 @@ hwloc__xml_import_diff_one(hwloc__xml_import_state_t state,
     }
   }
 
-  return state->close_tag(state);
+  return state->global->close_tag(state);
 }
 
 int
@@ -720,7 +844,7 @@ hwloc__xml_import_diff(hwloc__xml_import_state_t state,
     char *tag;
     int ret;
 
-    ret = state->find_child(state, &childstate, &tag);
+    ret = state->global->find_child(state, &childstate, &tag);
     if (ret < 0)
       return -1;
     if (!ret)
@@ -734,7 +858,7 @@ hwloc__xml_import_diff(hwloc__xml_import_state_t state,
     if (ret < 0)
       return ret;
 
-    state->close_child(&childstate);
+    state->global->close_child(&childstate);
   }
 
   *firstdiffp = firstdiff;
@@ -747,7 +871,8 @@ hwloc__xml_import_diff(hwloc__xml_import_state_t state,
 
 static int
 hwloc_xml__handle_distances(struct hwloc_topology *topology,
-			    struct hwloc_xml_backend_data_s *data)
+			    struct hwloc_xml_backend_data_s *data,
+			    const char *msgprefix)
 {
   struct hwloc_xml_imported_distances_s *xmldist, *next = data->first_distances;
 
@@ -766,8 +891,8 @@ hwloc_xml__handle_distances(struct hwloc_topology *topology,
     if (nbobjs != xmldist->distances.nbobjs) {
       /* distances invalid, drop */
       if (hwloc__xml_verbose())
-	fprintf(stderr, "ignoring invalid distance matrix with %u objs instead of %u\n",
-		xmldist->distances.nbobjs, nbobjs);
+	fprintf(stderr, "%s: ignoring invalid distance matrix with %u objs instead of %u\n",
+		msgprefix, xmldist->distances.nbobjs, nbobjs);
       free(xmldist->distances.latency);
     } else {
       /* distances valid, add it to the internal OS distances list for grouping */
@@ -803,6 +928,8 @@ hwloc_look_xml(struct hwloc_backend *backend)
   hwloc_localeswitch_declare;
   int ret;
 
+  state.global = data;
+
   assert(!topology->levels[0][0]->cpuset);
 
   hwloc_localeswitch_init();
@@ -814,22 +941,22 @@ hwloc_look_xml(struct hwloc_backend *backend)
     goto failed;
 
   /* find root object tag and import it */
-  ret = state.find_child(&state, &childstate, &tag);
+  ret = state.global->find_child(&state, &childstate, &tag);
   if (ret < 0 || !ret || strcmp(tag, "object"))
     goto failed;
   ret = hwloc__xml_import_object(topology, data, topology->levels[0][0], &childstate);
   if (ret < 0)
     goto failed;
-  state.close_child(&childstate);
+  state.global->close_child(&childstate);
 
   /* find end of topology tag */
-  state.close_tag(&state);
+  state.global->close_tag(&state);
 
   /* keep the "Backend" information intact */
   /* we could add "BackendSource=XML" to notify that XML was used between the actual backend and here */
 
   /* if we added some distances, we must check them, and make them groupable */
-  if (hwloc_xml__handle_distances(topology, data) < 0)
+  if (hwloc_xml__handle_distances(topology, data, data->msgprefix) < 0)
     goto err;
   data->first_distances = data->last_distances = NULL;
   topology->support.discovery->pu = 1;
@@ -841,7 +968,8 @@ hwloc_look_xml(struct hwloc_backend *backend)
   if (data->look_failed)
     data->look_failed(data);
   if (hwloc__xml_verbose())
-    fprintf(stderr, "XML component discovery failed.\n");
+    fprintf(stderr, "%s: XML component discovery failed.\n",
+	    data->msgprefix);
  err:
   hwloc_localeswitch_fini();
   return -1;
@@ -853,11 +981,22 @@ hwloc_topology_diff_load_xml(hwloc_topology_t topology __hwloc_attribute_unused,
 			     const char *xmlpath,
 			     hwloc_topology_diff_t *firstdiffp, char **refnamep)
 {
+  struct hwloc__xml_import_state_s state;
+  struct hwloc_xml_backend_data_s fakedata; /* only for storing global info during parsing */
   hwloc_localeswitch_declare;
-  char *env;
+  const char *basename;
   int force_nolibxml;
   int ret;
 
+  state.global = &fakedata;
+
+  basename = strrchr(xmlpath, '/');
+  if (basename)
+    basename++;
+  else
+    basename = xmlpath;
+  fakedata.msgprefix = strdup(basename);
+
   if (!hwloc_libxml_callbacks && !hwloc_nolibxml_callbacks) {
     errno = ENOSYS;
     return -1;
@@ -867,13 +1006,12 @@ hwloc_topology_diff_load_xml(hwloc_topology_t topology __hwloc_attribute_unused,
 
   *firstdiffp = NULL;
 
-  env = getenv("HWLOC_NO_LIBXML_IMPORT");
-  force_nolibxml = (env && atoi(env));
+  force_nolibxml = hwloc_nolibxml_import();
 retry:
   if (!hwloc_libxml_callbacks || (hwloc_nolibxml_callbacks && force_nolibxml))
-    ret = hwloc_nolibxml_callbacks->import_diff(xmlpath, NULL, 0, firstdiffp, refnamep);
+    ret = hwloc_nolibxml_callbacks->import_diff(&state, xmlpath, NULL, 0, firstdiffp, refnamep);
   else {
-    ret = hwloc_libxml_callbacks->import_diff(xmlpath, NULL, 0, firstdiffp, refnamep);
+    ret = hwloc_libxml_callbacks->import_diff(&state, xmlpath, NULL, 0, firstdiffp, refnamep);
     if (ret < 0 && errno == ENOSYS) {
       hwloc_libxml_callbacks = NULL;
       goto retry;
@@ -881,6 +1019,8 @@ retry:
   }
 
   hwloc_localeswitch_fini();
+
+  free(fakedata.msgprefix);
   return ret;
 }
 
@@ -890,11 +1030,15 @@ hwloc_topology_diff_load_xmlbuffer(hwloc_topology_t topology __hwloc_attribute_u
 				   const char *xmlbuffer, int buflen,
 				   hwloc_topology_diff_t *firstdiffp, char **refnamep)
 {
+  struct hwloc__xml_import_state_s state;
+  struct hwloc_xml_backend_data_s fakedata; /* only for storing global info during parsing */
   hwloc_localeswitch_declare;
-  char *env;
   int force_nolibxml;
   int ret;
 
+  state.global = &fakedata;
+  fakedata.msgprefix = "xmldiffbuffer";
+
   if (!hwloc_libxml_callbacks && !hwloc_nolibxml_callbacks) {
     errno = ENOSYS;
     return -1;
@@ -904,13 +1048,12 @@ hwloc_topology_diff_load_xmlbuffer(hwloc_topology_t topology __hwloc_attribute_u
 
   *firstdiffp = NULL;
 
-  env = getenv("HWLOC_NO_LIBXML_IMPORT");
-  force_nolibxml = (env && atoi(env));
-retry:
+  force_nolibxml = hwloc_nolibxml_import();
+ retry:
   if (!hwloc_libxml_callbacks || (hwloc_nolibxml_callbacks && force_nolibxml))
-    ret = hwloc_nolibxml_callbacks->import_diff(NULL, xmlbuffer, buflen, firstdiffp, refnamep);
+    ret = hwloc_nolibxml_callbacks->import_diff(&state, NULL, xmlbuffer, buflen, firstdiffp, refnamep);
   else {
-    ret = hwloc_libxml_callbacks->import_diff(NULL, xmlbuffer, buflen, firstdiffp, refnamep);
+    ret = hwloc_libxml_callbacks->import_diff(&state, NULL, xmlbuffer, buflen, firstdiffp, refnamep);
     if (ret < 0 && errno == ENOSYS) {
       hwloc_libxml_callbacks = NULL;
       goto retry;
@@ -1188,7 +1331,6 @@ hwloc__xml_export_diff(hwloc__xml_export_state_t parentstate, hwloc_topology_dif
 int hwloc_topology_export_xml(hwloc_topology_t topology, const char *filename)
 {
   hwloc_localeswitch_declare;
-  char *env;
   int force_nolibxml;
   int ret;
 
@@ -1199,8 +1341,7 @@ int hwloc_topology_export_xml(hwloc_topology_t topology, const char *filename)
 
   hwloc_localeswitch_init();
 
-  env = getenv("HWLOC_NO_LIBXML_EXPORT");
-  force_nolibxml = (env && atoi(env));
+  force_nolibxml = hwloc_nolibxml_export();
 retry:
   if (!hwloc_libxml_callbacks || (hwloc_nolibxml_callbacks && force_nolibxml))
     ret = hwloc_nolibxml_callbacks->export_file(topology, filename);
@@ -1220,7 +1361,6 @@ retry:
 int hwloc_topology_export_xmlbuffer(hwloc_topology_t topology, char **xmlbuffer, int *buflen)
 {
   hwloc_localeswitch_declare;
-  char *env;
   int force_nolibxml;
   int ret;
 
@@ -1231,8 +1371,7 @@ int hwloc_topology_export_xmlbuffer(hwloc_topology_t topology, char **xmlbuffer,
 
   hwloc_localeswitch_init();
 
-  env = getenv("HWLOC_NO_LIBXML_EXPORT");
-  force_nolibxml = (env && atoi(env));
+  force_nolibxml = hwloc_nolibxml_export();
 retry:
   if (!hwloc_libxml_callbacks || (hwloc_nolibxml_callbacks && force_nolibxml))
     ret = hwloc_nolibxml_callbacks->export_buffer(topology, xmlbuffer, buflen);
@@ -1256,7 +1395,6 @@ hwloc_topology_diff_export_xml(hwloc_topology_t topology __hwloc_attribute_unuse
 {
   hwloc_localeswitch_declare;
   hwloc_topology_diff_t tmpdiff;
-  char *env;
   int force_nolibxml;
   int ret;
 
@@ -1276,8 +1414,7 @@ hwloc_topology_diff_export_xml(hwloc_topology_t topology __hwloc_attribute_unuse
 
   hwloc_localeswitch_init();
 
-  env = getenv("HWLOC_NO_LIBXML_EXPORT");
-  force_nolibxml = (env && atoi(env));
+  force_nolibxml = hwloc_nolibxml_export();
 retry:
   if (!hwloc_libxml_callbacks || (hwloc_nolibxml_callbacks && force_nolibxml))
     ret = hwloc_nolibxml_callbacks->export_diff_file(diff, refname, filename);
@@ -1301,7 +1438,6 @@ hwloc_topology_diff_export_xmlbuffer(hwloc_topology_t topology __hwloc_attribute
 {
   hwloc_localeswitch_declare;
   hwloc_topology_diff_t tmpdiff;
-  char *env;
   int force_nolibxml;
   int ret;
 
@@ -1321,8 +1457,7 @@ hwloc_topology_diff_export_xmlbuffer(hwloc_topology_t topology __hwloc_attribute
 
   hwloc_localeswitch_init();
 
-  env = getenv("HWLOC_NO_LIBXML_EXPORT");
-  force_nolibxml = (env && atoi(env));
+  force_nolibxml = hwloc_nolibxml_export();
 retry:
   if (!hwloc_libxml_callbacks || (hwloc_nolibxml_callbacks && force_nolibxml))
     ret = hwloc_nolibxml_callbacks->export_diff_buffer(diff, refname, xmlbuffer, buflen);
@@ -1340,7 +1475,6 @@ retry:
 
 void hwloc_free_xmlbuffer(hwloc_topology_t topology __hwloc_attribute_unused, char *xmlbuffer)
 {
-  char *env;
   int force_nolibxml;
 
   if (!hwloc_libxml_callbacks && !hwloc_nolibxml_callbacks) {
@@ -1348,8 +1482,7 @@ void hwloc_free_xmlbuffer(hwloc_topology_t topology __hwloc_attribute_unused, ch
     return ;
   }
 
-  env = getenv("HWLOC_NO_LIBXML_EXPORT");
-  force_nolibxml = (env && atoi(env));
+  force_nolibxml = hwloc_nolibxml_export();
   if (!hwloc_libxml_callbacks || (hwloc_nolibxml_callbacks && force_nolibxml))
     hwloc_nolibxml_callbacks->free_buffer(xmlbuffer);
   else
@@ -1405,7 +1538,7 @@ hwloc_export_obj_userdata_base64(void *reserved,
   hwloc__xml_export_state_t state = reserved;
   size_t encoded_length;
   char *encoded_buffer;
-  int ret;
+  int ret __hwloc_attribute_unused;
 
   if (name && hwloc__xml_export_check_buffer(name, strlen(name)) < 0) {
     errno = EINVAL;
@@ -1444,6 +1577,7 @@ hwloc_xml_backend_disable(struct hwloc_backend *backend)
 {
   struct hwloc_xml_backend_data_s *data = backend->private_data;
   data->backend_exit(data);
+  free(data->msgprefix);
   free(data);
 }
 
@@ -1455,11 +1589,11 @@ hwloc_xml_component_instantiate(struct hwloc_disc_component *component,
 {
   struct hwloc_xml_backend_data_s *data;
   struct hwloc_backend *backend;
-  char *env;
   int force_nolibxml;
   const char * xmlpath = (const char *) _data1;
   const char * xmlbuffer = (const char *) _data2;
   int xmlbuflen = (int)(uintptr_t) _data3;
+  const char *basename;
   int err;
 
   if (!hwloc_libxml_callbacks && !hwloc_nolibxml_callbacks) {
@@ -1487,8 +1621,18 @@ hwloc_xml_component_instantiate(struct hwloc_disc_component *component,
   backend->disable = hwloc_xml_backend_disable;
   backend->is_thissystem = 0;
 
-  env = getenv("HWLOC_NO_LIBXML_IMPORT");
-  force_nolibxml = (env && atoi(env));
+  if (xmlpath) {
+    basename = strrchr(xmlpath, '/');
+    if (basename)
+      basename++;
+    else
+      basename = xmlpath;
+  } else {
+    basename = "xmlbuffer";
+  }
+  data->msgprefix = strdup(basename);
+
+  force_nolibxml = hwloc_nolibxml_import();
 retry:
   if (!hwloc_libxml_callbacks || (hwloc_nolibxml_callbacks && force_nolibxml))
     err = hwloc_nolibxml_callbacks->backend_init(data, xmlpath, xmlbuffer, xmlbuflen);
@@ -1505,6 +1649,7 @@ retry:
   return backend;
 
  out_with_data:
+  free(data->msgprefix);
   free(data);
  out_with_backend:
   free(backend);
@@ -1523,6 +1668,7 @@ static struct hwloc_disc_component hwloc_xml_disc_component = {
 
 const struct hwloc_component hwloc_xml_component = {
   HWLOC_COMPONENT_ABI,
+  NULL, NULL,
   HWLOC_COMPONENT_TYPE_DISC,
   0,
   &hwloc_xml_disc_component
